# 포켓 큐브
from .core import *


class poketCube( Cube ) :
    """
    포켓큐브(2*2*2) 게임
    """

    def __init__( self ) :
        self.size = 2
        self.make( self.size )
        self.history = [ ]
        self.done = None
        self.point = None
        self.count = 0
        self.set = ('F', 'F`', 'R', 'R`', 'U', 'U`', 'B', 'B`', 'L', 'L`', 'D', 'D`')

        # todo:랜덤한 값을 대입

        self.check( )

    def __repr__( self ) :
        """
        아래와 같은 구조르 출력됨
        ex) 루빅스큐브 출력시
        |================================|
        |  포켓 큐브 게임 2*2*2             |
        |       -------  완료 여부 : False |
        |       | 1 2 |  점수 : 00        |
        |       | 1 2 |  회전횟수 : 00     |
        | -------------------------      |
        | | 1 2 | 1 2 | 1 2 | 1 2 |      |
        | | 1 2 | 1 2 | 1 2 | 1 2 |      |
        | -------------------------      |
        |       | 1 2 |                  |
        |       | 1 2 |                  |
        |       -------                  |
        |================================|
        기록 :


        :return:
        """
        f1, f2, f3, f4, f5, f6 = self.cube[ 1 ].matrix, self.cube[ 2 ].matrix, self.cube[ 3 ].matrix, self.cube[
            4 ].matrix, self.cube[ 5 ].matrix, self.cube[ 6 ].matrix

        # 큐브 사이즈에 맞게 자동 조절
        # 단일 객체 길이 측정
        sample = len( '{}'.format( f1[ 0 ][ 0 ] ) )
        nullArea = ' ' * (5 + sample * 2)
        stick1Area = '-' * (5 + sample * 2)
        stick2Area = ' ' + '-' * (17 + sample * 8) + '\n'
        str = '=======================================\n'
        str += '  포켓 큐브 게임 2*2*2\n\n'
        str += '{}{}    완료 여부 : {done}\n'.format( nullArea, stick1Area, done=self.done )
        str += '{}| {} {} |    점수      : {point}\n'.format( nullArea, f2[ 0 ][ 0 ], f2[ 0 ][ 1 ], point=self.point )
        str += '{}| {} {} |    회전횟수  : {count}\n'.format( nullArea, f2[ 1 ][ 0 ], f2[ 1 ][ 1 ], count=self.count )
        str += stick2Area
        str += ' | {} {} | {} {} | {} {} | {} {} |\n'.format( f3[ 0 ][ 0 ], f3[ 0 ][ 1 ], f1[ 0 ][ 0 ], f1[ 0 ][ 1 ],
                                                              f4[ 0 ][ 0 ], f4[ 0 ][ 1 ], f6[ 0 ][ 0 ], f6[ 0 ][ 1 ] )
        str += ' | {} {} | {} {} | {} {} | {} {} |\n'.format( f3[ 1 ][ 0 ], f3[ 1 ][ 1 ], f1[ 1 ][ 0 ], f1[ 1 ][ 1 ],
                                                              f4[ 1 ][ 0 ], f4[ 1 ][ 1 ], f6[ 1 ][ 0 ], f6[ 1 ][ 1 ] )
        str += stick2Area
        str += '{}| {} {} |\n'.format( nullArea, f5[ 0 ][ 0 ], f5[ 0 ][ 1 ] )
        str += '{}| {} {} |\n'.format( nullArea, f5[ 1 ][ 0 ], f5[ 1 ][ 1 ] )
        str += '{}{}\n'.format( nullArea, stick1Area )
        str += ' 기록 : {}\n'.format( self.history )
        str += '====================================='
        return str

    def copy( self, faces ) :
        """
        딕셔너리에 적혀 있는대로 면을 복사해 준다
        :param faces: 키는 면의 인덱스, 값은 복사할 값이 있는 인덱스
        :return:

        faces = {1:'r1', 2: 'c0}
        dic = self.copy(faces)
        dic == {1:<1번면의 r1 값>, 2: <2번면의 c0 값>}
        """
        dic = faces
        assert type( dic ) == type( dict( ) )

        for i in dic :
            # 면에서 원하는 인덱스 값을 복사하여 알려준다
            dic[ i ] = self.cube[ i ].get( dic[ i ] )
        return dic

    # todo: 게임 상태를 알려주는 메소드 추가하기, 해당 메소드 실행시 튜블에 완성여부,점수,개별면 상태를 담아 반환한다
    # todo: action 메소드 실행시 큐브상태를 반환하도록 변경

    @checkDouble
    def rotate( self, action ) :
        """
        큐브를 회전시키는 메소드
        :param act: 명령어셋
        :return:
        """
        act = action
        """
        포켓큐브 명령어 셋
        가능한 명령 : F,F`,R,R`,U,U`,B,B`,L,L`,D,D`,F2,R2,U2,B2,L2,D2
        각 명령어에는 명령어 실행시 변경되는 정보를 딕셔너리로 담고 있다.

        rotate : 명령어 실행시 회전하는 면과 회전되는 방향
                num : 회전할 면 번호
                direction : 회전하는 방향

        old : 변경전 상태 캡쳐
                형식 -> key: 면 번호, value: 캡쳐할 부분(바뀌는 부분)
        change : 변경될 부분에 대한 정보
                형식 -> key: 변경될 부분의 면, value: 변경할 부분의 면 (면 번호,변경될 부분, 자료 반전 여부/True: 변경전 자료 반전시키기)

        """
        turnset = {
            'F' : {  # 완료
                'rotate' : {
                    'num' : 1,
                    'direction' : 'r',
                },
                'old' : {
                    2 : 'c1',
                    3 : 'r1',
                    4 : 'r0',
                    5 : 'c0'
                },
                'change' : {
                    2 : { 'num' : 4, 'index' : 'r0', 'flip' : False },
                    3 : { 'num' : 2, 'index' : 'c1', 'flip' : True },
                    4 : { 'num' : 5, 'index' : 'c0', 'flip' : True },
                    5 : { 'num' : 3, 'index' : 'r1', 'flip' : False },
                }
            },
            'F`' : {  # 완료
                'rotate' : {
                    'num' : 1,
                    'direction' : 'l',
                },
                'old' : {
                    2 : 'c1',
                    3 : 'r1',
                    4 : 'r0',
                    5 : 'c0'
                },
                'change' : {
                    2 : { 'num' : 3, 'index' : 'r1', 'flip' : True },
                    3 : { 'num' : 5, 'index' : 'c0', 'flip' : False },
                    4 : { 'num' : 2, 'index' : 'c1', 'flip' : False },
                    5 : { 'num' : 4, 'index' : 'r0', 'flip' : True },
                }
            },
            'R' : {  # 완료
                'rotate' : {
                    'num' : 4,
                    'direction' : 'r',
                },
                'old' : {
                    1 : 'r1',
                    2 : 'r1',
                    5 : 'r1',
                    6 : 'r0'
                },
                'change' : {
                    1 : { 'num' : 2, 'index' : 'r1', 'flip' : False },
                    2 : { 'num' : 6, 'index' : 'r0', 'flip' : True },
                    5 : { 'num' : 1, 'index' : 'r1', 'flip' : False },
                    6 : { 'num' : 5, 'index' : 'r1', 'flip' : True },
                }
            },
            'R`' : {  # 완료
                'rotate' : {
                    'num' : 4,
                    'direction' : 'l',
                },
                'old' : {
                    1 : 'r1',
                    2 : 'r1',
                    5 : 'r1',
                    6 : 'r0'
                },
                'change' : {
                    1 : { 'num' : 5, 'index' : 'r1', 'flip' : False },
                    2 : { 'num' : 1, 'index' : 'r1', 'flip' : False },
                    5 : { 'num' : 6, 'index' : 'r0', 'flip' : True },
                    6 : { 'num' : 2, 'index' : 'r1', 'flip' : True },
                }
            },
            'U' : {  # 완료
                'rotate' : {
                    'num' : 2,
                    'direction' : 'r',
                },
                'old' : {
                    1 : 'c0',
                    3 : 'c0',
                    4 : 'c0',
                    6 : 'c0'
                },
                'change' : {
                    1 : { 'num' : 3, 'index' : 'c0', 'flip' : False },
                    3 : { 'num' : 6, 'index' : 'c0', 'flip' : False },
                    4 : { 'num' : 1, 'index' : 'c0', 'flip' : False },
                    6 : { 'num' : 4, 'index' : 'c0', 'flip' : False },
                }
            },
            'U`' : {  # 완료
                'rotate' : {
                    'num' : 2,
                    'direction' : 'l',
                },
                'old' : {
                    1 : 'c0',
                    3 : 'c0',
                    4 : 'c0',
                    6 : 'c0'
                },
                'change' : {
                    1 : { 'num' : 4, 'index' : 'c0', 'flip' : False },
                    3 : { 'num' : 1, 'index' : 'c0', 'flip' : False },
                    4 : { 'num' : 6, 'index' : 'c0', 'flip' : False },
                    6 : { 'num' : 3, 'index' : 'c0', 'flip' : False },
                }
            },
            'B' : {  # 완료
                'rotate' : {
                    'num' : 6,
                    'direction' : 'r',
                },
                'old' : {
                    2 : 'c0',
                    3 : 'r0',
                    4 : 'r1',
                    5 : 'c1'
                },
                'change' : {
                    2 : { 'num' : 3, 'index' : 'r0', 'flip' : True },
                    3 : { 'num' : 5, 'index' : 'c1', 'flip' : False },
                    4 : { 'num' : 2, 'index' : 'c0', 'flip' : False },
                    5 : { 'num' : 4, 'index' : 'r1', 'flip' : True },
                }
            },
            'B`' : {  # 완료
                'rotate' : {
                    'num' : 6,
                    'direction' : 'l',
                },
                'old' : {
                    2 : 'c0',
                    3 : 'r0',
                    4 : 'r1',
                    5 : 'c1'
                },
                'change' : {
                    2 : { 'num' : 4, 'index' : 'r1', 'flip' : False },
                    3 : { 'num' : 2, 'index' : 'c0', 'flip' : True },
                    4 : { 'num' : 5, 'index' : 'c1', 'flip' : True },
                    5 : { 'num' : 3, 'index' : 'r0', 'flip' : False },
                }
            },
            'L' : {  # 완료
                'rotate' : {
                    'num' : 3,
                    'direction' : 'r',
                },
                'old' : {
                    1 : 'r0',
                    2 : 'r0',
                    5 : 'r0',
                    6 : 'r1'
                },
                'change' : {
                    1 : { 'num' : 5, 'index' : 'r0', 'flip' : False },
                    2 : { 'num' : 1, 'index' : 'r0', 'flip' : False },
                    5 : { 'num' : 6, 'index' : 'r1', 'flip' : True },
                    6 : { 'num' : 2, 'index' : 'r0', 'flip' : True },
                }
            },
            'L`' : {  # 완료
                'rotate' : {
                    'num' : 3,
                    'direction' : 'l',
                },
                'old' : {
                    1 : 'r0',
                    2 : 'r0',
                    5 : 'r0',
                    6 : 'r1'
                },
                'change' : {
                    1 : { 'num' : 2, 'index' : 'r0', 'flip' : False },
                    2 : { 'num' : 6, 'index' : 'r1', 'flip' : True },
                    5 : { 'num' : 1, 'index' : 'r0', 'flip' : False },
                    6 : { 'num' : 5, 'index' : 'r0', 'flip' : True },
                }
            },
            'D' : {  # 완료
                'rotate' : {
                    'num' : 5,
                    'direction' : 'r',
                },
                'old' : {
                    1 : 'c1',
                    3 : 'c1',
                    4 : 'c1',
                    6 : 'c1'
                },
                'change' : {
                    1 : { 'num' : 4, 'index' : 'c1', 'flip' : False },
                    3 : { 'num' : 1, 'index' : 'c1', 'flip' : False },
                    4 : { 'num' : 6, 'index' : 'c1', 'flip' : False },
                    6 : { 'num' : 3, 'index' : 'c1', 'flip' : False },
                }
            },
            'D`' : {  # 완료
                'rotate' : {
                    'num' : 5,
                    'direction' : 'l',
                },
                'old' : {
                    1 : 'c1',
                    3 : 'c1',
                    4 : 'c1',
                    6 : 'c1'
                },
                'change' : {
                    1 : { 'num' : 3, 'index' : 'c1', 'flip' : False },
                    3 : { 'num' : 6, 'index' : 'c1', 'flip' : False },
                    4 : { 'num' : 1, 'index' : 'c1', 'flip' : False },
                    6 : { 'num' : 4, 'index' : 'c1', 'flip' : False },
                }
            },
        }
        # 명령어셋 확인
        assert act in turnset

        ## 명령어셋 호출
        act = turnset[ act ]
        rotate = act[ 'rotate' ]  # 회전할 면
        old = act[ 'old' ]  # 변경될 영역
        change = act[ 'change' ]  # 변경할 영역

        ## 면 회전
        self.cube[ rotate[ 'num' ] ].rotate( rotate[ 'direction' ] )

        ## 면 변경
        # 면 변경을 위해 이전 면 상태 캡쳐
        old = self.copy( old )

        for oldnum in old :
            # 변경될 자료
            data = old[ oldnum ]
            # 변경할 면과 위치
            new = change[ oldnum ]

            # 자료 변경전 자료를 반전 시키기
            if new[ 'flip' ] :
                data = np.flip( data, 0 )

            # 자료 변경
            self.cube[ new[ 'num' ] ].change( new[ 'index' ], data )



import numpy as np
from functools import wraps
import random

"""
큐브 게임을 위한 핵식 클래스 모음
"""


class face :
    """
    큐브의 면을 표현 하는 객체
    큐브의 색(숫자)을 바꿔준다.
    """

    def __init__( self, n ) :
        """
        n*n 크기의 면의 만든다
        :param n:
        :return:
        """
        dtype = np.uint8
        self.size = n
        self.matrix = np.zeros( (self.size, self.size), dtype=dtype )

    def __repr__( self ) :

        return "\n" + str( self.matrix ) + "\n"

    def reset( self, num ) :
        """
        면의 모든 숫자를 num으로 바꿔준다
        큐브 초기 생성시 면 값을 리셋 시키기 위한 함수
        :param num:
        :return:
        """
        self.matrix[ :, : ] = num
        self.check( )

    def set( self, mat ) :
        """
        면의 초기값을 mat로 바꿔준다
        :param mat: ndarray 혹은 차원이 동일한 리스트로 만든 행렬
        :return:
        """
        newmat = mat

        # ndarray 형식 일 경우
        if type( newmat ) == type( self.matrix ) :
            # mat의 자료형 비교
            if not newmat.dtype == self.matrix.dtype :
                # 자료형이 맞지 않을 경우 face객체의 행렬과 자료형을 동일하게 변경
                newmat.dtype = self.matrix.dtype

        # 리스트 형식의 행렬일 경우
        elif type( newmat ) == type( list( ) ) :
            # 리스트를 ndarray 타입으로 변경
            newmat = np.array( newmat, dtype=self.matrix.dtype )

        # mat의 shape 비교
        assert newmat.shape == self.matrix.shape

        self.matrix = newmat
        self.check( )

    def check( self ) :
        """
        면의 상태정보를 갱신해 준다. 자료가 변경될때 이 메소드를 호출해 주면 자동으로 갱신한다.
        self.done은 면의 완성여부를 self.point는 면의 점수를 알려준다.
        점수는 면에 가장 많이 중복된 숫자의 갯수로 매겨진다.
        루빅스 큐브를 예로 들면 한 면에서 최대 9점이 나올수 있으며 루빅스 큐브 한개에서 최대 54점이 나온다.
        :return:
        """

        num, count = np.unique( self.matrix, return_counts=True )

        # 한면에 있는 숫자와 갯수
        # ex) [[1,2,1],[2,3,1],[2,4,1]] 일 경우 -> self.status == [(1, 4), (2, 3), (3, 1), (4, 1)]
        self.status = list( zip( num, count ) )
        if len( self.status ) == 1 :  # 한면의 모든 숫자가 동일할 경우
            self.done = True
        else :
            self.done = False

        self.point = self.status[ 0 ][ 1 ]

    """
    액션 메소드
    열(row)과 행(col), 인덱싱으로 구분하며 선택된 것은 1로 표시됨
        _____       _____
        |0|0|       |0|1|
        --+--  -->  --+--
        |0|0|       |0|1|
        -----       -----

        왼쪽에서 r1은 [:,:1] 이것과 같은 의미와 위에처럼 표시한다.
    """

    def get( self, index ) :
        """
        해당 인덱스의 값을 넘겨준다
        :param index: 열,행과 인덱스 번호로 이뤄진 텍스트 ex) "r0", "c4"
        :return:
        """
        indexcode = int( index[ 1 : ] )

        if index[ 0 ] == "r" :
            return np.copy( self.matrix[ :, indexcode ] )
        elif index[ 0 ] == "c" :
            return np.copy( self.matrix[ indexcode, : ] )

    def getface( self ) :
        """
        n*n 행열을 1차 행열로 넘겨줍니다.
        :return:
        """
        return np.reshape( self.matrix, (1, pow(self.size,2)) )

    def change( self, index, data ) :
        """
        입력받은 인덱스의 값을 바꿔준다
        :param index:
        :param data:
        :return:
        """

        ## 인덱싱문구 해석

        # 인덱싱 번호
        indexcode = int( index[ 1 : ] )

        if index[ 0 ] == "r" :
            self.matrix[ :, indexcode ] = data
        elif index[ 0 ] == "c" :
            self.matrix[ indexcode, : ] = data

        # 상태갱신
        self.check( )

    def rotate( self, Direction="r" ) :
        """
        면을 오른쪽 혹은 왼쪽으로 회전한다.
        기본은 오른쪽으로 회전
        :param count: (r :오른쪽으로 회전, l: 왼쪽으로 회전)
        :return:
        """
        # 방향 값 확인
        direc = Direction
        assert direc in ('r', 'l')

        if direc == 'r' :
            self.matrix = np.rot90( self.matrix, 3 )
        elif direc == 'l' :
            self.matrix = np.rot90( self.matrix )

        # 상태 갱신
        self.check( )


# 두번 반복 명령어 인식 데코레이터
def checkDouble( func ) :
    @wraps( func )
    def wrapper( *args ) :
        if args[ 1 ][ -1 ] == '2' :
            # 명령어 끝에 2가 있을 경우 해당 명령어 2번 반복
            act = args[ 1 ][ :-1 ]
            for _ in range( 2 ) :
                func( args[ 0 ], act )
        else :
            func( *args )

    return wrapper


class Cube :
    """
    큐브 생성 관리에 유용한 메소드 모음
    """

    def __init__( self ) :
        self.history = [ ]
        self.size = 0 # 큐브 크기
        self.done = None  # 큐브 완성여부
        self.reward = 0  # 큐브 점수
        self.point = 0
        self.count = 0  # 큐브 회전 횟수
        self.set = None  # 사용가능한 회전 명령어 모음
        self.scram = []  # 사용된 스크램블
        self.faces = None  # 기계학습에 활용될 면 상태 (6*9)

    def reset( self ) :
        """
        큐브 초기화
        :return:
        """
        self.history = [ ]
        self.point = 0
        self.reward = 0
        self.scram = []
        for i in range( 1, 7 ) :
            self.cube[ i ].reset( i )
        self.check( )

    def make( self, n ) :
        """
        큐브는 딕셔너리(ey = 면의 인덱스 번호, value = 면 인스턴스)로 구성되있다.
        면에 인덱스 번호를 부여하는 방식은 READMD.md파일의 큐브 배치도를 참고 바람.
        모든 행동은 1번 면이 가장 앞에 있는 상태를 기준으로 이뤄진다.
        :param n: 생성할 큐브 차원
        :return:
        """
        size = n

        self.cube = { }
        for i in range( 1, 7 ) :
            # 면을 생성한다.
            self.cube[ i ] = face( size )
            # 면의 값을 초기화 시킨다 ex) 3번 면의 값은 모두 3으로 초기화
            self.cube[ i ].reset( i )

    def check( self ) :
        """
        면 상태 체크 메소드, 큐브가 변경되는 시점마다 호출하여 완셩여부와 점수를 확인한다.
        :return:
        """
        # 큐브 완셩 여부 확인
        done = [ self.cube[ x ].done for x in self.cube ]
        if False in done :
            self.done = False
        else :
            self.done = True

        # 회전 횟수
        self.count = len( self.history )

        # 점수 갱신
        reward = -1
        if self.count == 0:
            # 게임 시작 전일 경우 보상은 0으로 한다
            self.reward = 0
        elif True in done:
            # 완성된 면이 존재할 경우 완성된 면의 갯수*면의 총점수 만큼 보상 부여
            self.reward = done.count(True)*pow(self.size,2)
        else:
            # 큐브가 미완성일경우 점수 차감
            self.reward = -1




    def __repr__( self ) :
        """
        아래와 같은 구조르 출력됨
        ex) 루빅스큐브 출력시
        |=====================================|
        |  3*3*3 큐브 게임                      |
        |         ---------                   |
        |         | 1 2 3 |  완료 여부 : False  |
        |         | 1 2 3 |  점수 : 00         |
        |         | 1 2 3 |  회전횟수 : 00      |
        | ---------------------------------   |
        | | 1 2 3 | 1 2 3 | 1 2 3 | 1 2 3 |   |
        | | 1 2 3 | 1 2 3 | 1 2 3 | 1 2 3 |   |
        | | 1 2 3 | 1 2 3 | 1 2 3 | 1 2 3 |   |
        | ---------------------------------   |
        |         ---------                   |
        |         | 1 2 3 |                   |
        |         | 1 2 3 |                   |
        |         | 1 2 3 |                   |
        |         ---------                   |
        |=====================================|
        기록 :


        :return:
        """

    @checkDouble  # 반복명령어 처리 데코레이터
    def rotate( self, action ) :
        """
        입력받은 회전방향으로 큐브를 돌린다.
        각 큐브 게임에서 오버라이팅해서 사용하면 됨
        또한 반복명령어(ex. F2, F`2)를 처리하고 싶으면 @checkDouble 데코레이터를 달아주면 된다
        :param action:
        :return:
        """
        pass

    def getcube(self):
        faces = None
        for i in range( 1, 7 ) :
            face = self.cube[ i ]
            if i == 1 :
                faces = face.getface( )
            else :
                faces = np.append( faces, face.getface( ), axis=0 )
        return faces

    def action( self, action ) :
        """
        입력받은 act를 수행한뒤 상태를 반환해준다
        :param act: 회전 방향 기호
        :return: (완료여부,점수,회전횟수,큐브화면)
        """
        self.rotate( action )
        # 회전 기록
        self.history.append( action )
        # todo:180되 회전 명령어대신 90도 명령어가 2번 표기 되도록 변경
        # 상태 갱신
        self.check( )
        self.point += self.reward

        return (self.done, self.reward, self.count, self.getcube())

    def scramble( self, len=25, count=5, hide =True ) :
        """
        램덤한 5개의 스크램블을 생성한뒤 램덤으로 하나의 스크램블을 선택하여 큐브 모양을 만들어 준다.
        :param len: 스크램블 길이
        :param count: 생성할 스크램블 갯수
        :param hide: 스크램블을 출력할지 여부, 출력여부와 상관없이 self.scramble에 저장은 된다.
        :return:
        """
        # 큐브 초기화
        self.reset( )

        # 임의의 스크램블 모음
        scrambles = [ ]
        # 사용가능한 명령어 모음
        set = self.set

        lenth = len
        num = count

        def mix( ) :
            """
            스크램블을 만든다
            :return:
            """
            scramble = [ ]
            for _ in range( lenth ) :
                # 스크램블에 임의의 명령어 추가
                scramble.append( random.choice( set ) )
            # 완성된 스크램블 순서를 뒤섞기
            random.shuffle( scramble )
            return scramble

        for _ in range( num ) :
            scrambles.append( mix( ) )

        # 스크램블 선택
        self.scram = random.choice( scrambles )

        # 스크램블 하기
        for r in self.scram :
            self.rotate( r )
        self.check( )
        # 스크램블 순서를 보여준다
        if not hide:
            print( "scramble by ", self.scram )

        return self.getcube()